///// HEADER FILE


#ifndef __PLAT_DBFILTER_H
#define __PLAT_DBFILTER_H

// =================================
// FILTERS by Frederic BATTY, 1995
// =================================

// This encapsulation allows one to create an IDAPI filter by a C++ statement
// like this :
//
// Cursor->SetFilter ( Field("FLD1") > 5 || Field("FLD2").IsBlank() )
//
//

#ifndef __REF_H
  #include <ref.h>  // class TReference
#endif

#ifndef __CSTRING_H
  #include  <cstring.h> // class string
#endif

#ifndef DBI_H
  #include  <idapi.h>
#endif

// modifier to #define in order to use these classes from a DLL.
#ifndef _SBASECLASS
#define _SBASECLASS
#endif

// assert macro
#ifndef SCHECK(x)
#define SCHECK(x)
#endif

// error handler
#ifndef SERRORX(x)
#define SERRORX(x)
#endif

// Date / Time classes. Can be replaced by your own, provided there exists a
// global 'ToIDAPI' function, that converts the class to IDAPI's internal format.
// To do this, #define DBIDATE, DBITIME, DBITIMESTAMP to the actual class names
// before including this header.
// By default, these classes are simple wrappers around the IDAPI types.
#ifndef DBIDATE
struct DBIDATE
{
    DATE date;
    DBIDATE(DATE d) : date(d) {}
};

struct DBITIME
{
    TIME time;
    DBITIME(TIME t) : time(t) {}
};

struct DBITIMESTAMP
{
    TIMESTAMP timestamp;
    DBITIMESTAMP(TIMESTAMP t) : timestamp(t) {}
};

inline DATE ToIDAPI(DBIDATE d) { return d.date; }
inline TIME ToIDAPI(DBITIME t) { return t.time; }
inline TIMESTAMP ToIDAPI(DBITIMESTAMP t) { return t.timestamp; }

#endif

// --- Reference counting pointer ------------------------------------------
//   'T' must be derived from TReference
// -------------------------------------------------------------------------

template <class T> class SRefPtr {

    T*          Ptr;    // Pointeur sur l'objet rfrenc

    void RemovePtr()
            { if(Ptr){ Ptr->RemoveReference(); if(Ptr->References()<=0) delete Ptr; } }
    void SetPtr( T* ptr )
            { Ptr=ptr; if(Ptr) {Ptr->AddReference();} }

  public:
    // Constructors & destructor
    // -------------------------
    SRefPtr( )            { Ptr = 0; }
    SRefPtr( T* ptr )     { SetPtr(ptr); }
    ~SRefPtr()            { RemovePtr(); }

    // Assignments operators
    // ---------------------
    SRefPtr<T>& operator =( T* ptr ) { if(ptr!=Ptr){ RemovePtr(); SetPtr(ptr); } return *this; }
    SRefPtr<T>& operator =( const SRefPtr<T>& t ) { if(Ptr!=t.Ptr){ RemovePtr(); SetPtr(t.Ptr); } return *this; }
    SRefPtr<T>& operator()( T* ptr ) { if(ptr!=Ptr){ RemovePtr(); SetPtr(ptr); } return *this; }
    SRefPtr<T>& operator()( const SRefPtr<T>& t ) { if(Ptr!=t.Ptr){ RemovePtr(); SetPtr(t.Ptr); } return *this; }

    // Comparison
    // ----------
    friend int operator ==( const SRefPtr<T>& p0, const SRefPtr<T>& p1 ) { return p0.Ptr==p1.Ptr; }
    friend int operator ==( const SRefPtr<T>& p0, const T* p1 )          { return p0.Ptr==p1; }
    friend int operator ==( const T* p0, const SRefPtr<T>& p1 )          { return p0==p1.Ptr; }

    // Automatic cast to a T*
    // ----------------------
    T*  operator() () const     { return Ptr;  }
    T*  Get() const             { return Ptr;  }
    operator T*() const         { return Ptr;   }
    T& operator *() const       { return *Ptr; }
    T* operator->() const       { return Ptr;   }
};

// Internal expression tree built by the compiler.

struct _SBASECLASS BaseNode : public TReference
{
    struct CompileStruct;

    // Translation to the IDAPI format (2 passes)
    virtual void Compile(CompileStruct&) = 0;

    virtual ~BaseNode() {}
};

struct _SBASECLASS UnaryNode : public BaseNode
{
    CANOp Oper;
    SRefPtr<BaseNode> Node1;

    UnaryNode( CANOp op, BaseNode* node1) :
        Oper(op), Node1(node1)
    {
        SCHECK(node1);
    }

    virtual void Compile(CompileStruct& data);
};

struct _SBASECLASS BinaryNode : public BaseNode
{
    CANOp Oper;
    SRefPtr<BaseNode> Node1, Node2;

    BinaryNode( CANOp op, BaseNode* node1, BaseNode* node2) :
        Oper(op), Node1(node1), Node2(node2)
    {
        SCHECK(node1);
        SCHECK(node2);
    }

    virtual void Compile(CompileStruct& data);

};

struct _SBASECLASS CompareNode : public BaseNode
{
    CANOp Oper;                     // canLIKE, canEQ 
    BOOL    CaseInsensitive;        // 3 val: UNKNOWN = "fastest", "native"
    UINT16  PartialLen;             // Partial fieldlength (0 is full length)
    SRefPtr<BaseNode> Node1, Node2;   
    CompareNode( CANOp op, BaseNode* node1, BaseNode* node2, BOOL bCaseInsensitive=0, UINT16 iPartialLen=0) :
        Oper(op), Node1(node1), Node2(node2),
        CaseInsensitive(bCaseInsensitive),
        PartialLen(iPartialLen)
    {
        SCHECK(node1);
        SCHECK(node2);
    }

    virtual void Compile(CompileStruct& data);
};

struct _SBASECLASS FieldNode : public BaseNode
{
    string FldName;
    FieldNode(const string& fldname) : FldName(fldname)
    {
        SCHECK(!fldname.is_null());
    }

    virtual void Compile(CompileStruct& data);
};

struct _SBASECLASS CstNode : public BaseNode
{
    UINT16 FldType;
    CstNode(UINT16 type) : FldType(type) {}

    virtual void        Compile(CompileStruct& data);
    virtual int         ValueSize() = 0;
    virtual const void* ValueAddr() = 0;
};

template<class T, UINT16 type> struct ConstantNode : public CstNode
{
    T   Value;
    ConstantNode(T val) : CstNode(type), Value(val) {}
    virtual int         ValueSize() { return sizeof(T); }
    virtual const void* ValueAddr() { return &Value; }
};

struct ConstantStringNode : public CstNode
{
    string  Value;
    ConstantStringNode(const string& val) : CstNode(fldZSTRING), Value(val) {}
    virtual int         ValueSize() { return Value.length() + 1; }
    virtual const void* ValueAddr() { return Value.c_str(); }
};

struct _SBASECLASS DBINode
{
    typedef const DBINode& CRNode;

    SRefPtr<BaseNode> _Node;

    BaseNode* Get() const { return _Node.Get(); }
    int     is_null() const { return _Node.Get()==0; }

    DBINode(const DBINode& n) : _Node(n.Get()) {}
    DBINode(BaseNode* n=0) : _Node(n) {}

    // to be eliminated when BOOL becomes a separate type
    // ( to use a bool in an expression, use the syntax DBINode::True or DBINode::False )
    enum Bool { False=FALSE, True=TRUE };

    DBINode(Bool val)                       : _Node(new ConstantNode<BOOL, fldBOOL>(val)) {}
    DBINode(INT16 val)                      : _Node(new ConstantNode<INT16, fldINT16>(val)) {}
    DBINode(INT32 val)                      : _Node(new ConstantNode<INT32, fldINT32>(val)) {}
    DBINode(FLOAT val)                      : _Node(new ConstantNode<FLOAT, fldFLOAT>(val)) {}
    DBINode(DBIDATE val)                    : _Node(new ConstantNode<DATE,  fldDATE> (ToIDAPI(val))) {}
    DBINode(DBITIME val)                    : _Node(new ConstantNode<TIME,  fldTIME> (ToIDAPI(val))) {}
    DBINode(DBITIMESTAMP val)               : _Node(new ConstantNode<TIMESTAMP,  fldTIMESTAMP> (ToIDAPI(val))) {}
    DBINode(const char* val)                : _Node(new ConstantStringNode(val)) {}
    DBINode(const string& val)              : _Node(new ConstantStringNode(val)) {}

    DBINode& operator= (Bool val)           { _Node = new ConstantNode<BOOL, fldBOOL>(val); return *this; }
    DBINode& operator= (INT16 val)          { _Node = new ConstantNode<INT16, fldINT16>(val); return *this; }
    DBINode& operator= (INT32 val)          { _Node = new ConstantNode<INT32, fldINT32>(val); return *this; }
    DBINode& operator= (FLOAT val)          { _Node = new ConstantNode<FLOAT, fldFLOAT>(val); return *this; }
    DBINode& operator= (DBIDATE val)        { _Node = new ConstantNode<DATE,  fldDATE> (ToIDAPI(val)); return *this; }
    DBINode& operator= (DBITIME val)        { _Node = new ConstantNode<TIME,  fldTIME> (ToIDAPI(val)); return *this; }
    DBINode& operator= (DBITIMESTAMP val)   { _Node = new ConstantNode<TIMESTAMP,  fldTIMESTAMP> (ToIDAPI(val)); return *this; }
    DBINode& operator= (const char* val)    { _Node = new ConstantStringNode(val); return *this; }
    DBINode& operator= (const string& val)  { _Node = new ConstantStringNode(val); return *this; }

    DBINode& operator= (BaseNode* n)        { _Node = n; return *this; }
    DBINode& operator= (CRNode n)           { _Node = n.Get(); return *this; }

    #define UNARYOP(Op, canOp)                                      \
    friend DBINode operator##Op (CRNode n1)                         \
    {                                                               \
        return DBINode(new UnaryNode(canOp, n1.Get()));             \
    }

    UNARYOP(!, canNOT);
    UNARYOP(-, canMINUS);

    DBINode IsBlank()
    {
        return DBINode(new UnaryNode(canISBLANK, Get()));
    }

    DBINode NotBlank()
    {
        return DBINode(new UnaryNode(canNOTBLANK, Get()));
    }

    #define BINARYOP(Op, canOp)                                     \
    friend DBINode operator##Op (CRNode n1, CRNode n2)              \
    {                                                               \
        return DBINode(new BinaryNode(canOp, n1.Get(), n2.Get()));  \
    }

    BINARYOP(+, canADD);
    BINARYOP(-, canSUB);
    BINARYOP(*, canMUL);
    BINARYOP(/, canDIV);
    BINARYOP(%, canMOD);
    BINARYOP(&&, canAND);
    BINARYOP(||, canOR);
    BINARYOP(==, canEQ);
    BINARYOP(!=, canNE);
    BINARYOP(>, canGT);
    BINARYOP(>=, canGE);
    BINARYOP(<, canLT);
    BINARYOP(<=, canLE);

    DBINode Equals(CRNode n2, BOOL bCaseInsensitive = FALSE, UINT16 iPartialLen = 0)
    {
        return DBINode(new CompareNode(canEQ, Get(), n2.Get(), bCaseInsensitive, iPartialLen));
    }

    DBINode Like(CRNode n2, BOOL bCaseInsensitive = FALSE, UINT16 iPartialLen = 0)
    {
        return DBINode(new CompareNode(canLIKE, Get(), n2.Get(), bCaseInsensitive, iPartialLen));
    }

    // Translation to the IDAPI format 
    CANExpr* Compile();
};

// Output of the compiled expression
ostream& operator << (ostream& os, const CANExpr& expr);

struct _SBASECLASS Field : public DBINode
{
    Field(const string& fldname) : DBINode(new FieldNode(fldname)) {}
};

#endif __PLAT_DBFILTER_H
